#!/usr/bin/env python3

# Copyright (C) 2025 Zero ASIC
# This code is licensed under Apache License 2.0 (see LICENSE for details)

import multiprocessing
import numpy as np
from switchboard import SbDut, UmiTxRx, UmiCmd, random_umi_packet, umi_loopback
from umi import lumi


def gen_mergeable_umi_packets(num_bytes, size, opcode_arr, bytes_per_tx=16):

    packets = []

    if (bytes_per_tx > 16):
        num_mergeable_packets = bytes_per_tx // 16
    else:
        num_mergeable_packets = 1

    for _ in range(num_bytes // bytes_per_tx):
        dstaddr = (2**8)*np.random.randint(2**32)
        srcaddr = (2**8)*np.random.randint(2**32)
        opcode = np.random.choice(opcode_arr)
        for _ in range(num_mergeable_packets):
            length = ((bytes_per_tx // num_mergeable_packets) >> size) - 1
            txp = random_umi_packet(opcode=opcode, dstaddr=dstaddr,
                                    srcaddr=srcaddr, size=size, len=length,
                                    eom=0, eof=0, max_bytes=16)
            packets.append(txp)
            cmd_len = (txp.cmd >> 8) & 0xFF
            cmd_size = (txp.cmd >> 5) & 0x07
            cmd_bytes = (2**cmd_size) * (cmd_len + 1)
            dstaddr = dstaddr + cmd_bytes
            srcaddr = srcaddr + cmd_bytes

    return packets


def main():

    multiprocessing.set_start_method('fork')

    dut = SbDut('testbench', default_main=True, trace=True)

    dut.use(lumi)

    # Add testbench
    dut.input('lumi/testbench/testbench_lumi_ready.sv', package='umi')

    # Verilator configuration
    dut.set('tool', 'verilator', 'task', 'compile', 'file', 'config', 'lumi/testbench/config.vlt',
            package='umi')

    dut.build()

    # launch the simulation
    dut.simulate(
        plusargs=[
            ('valid_mode', 1),
            ('ready_mode', 1)
        ]
    )

    packets_req = gen_mergeable_umi_packets(num_bytes=65536,
                                            size=0,
                                            opcode_arr=[UmiCmd.UMI_REQ_READ,
                                                        UmiCmd.UMI_REQ_WRITE,
                                                        UmiCmd.UMI_REQ_POSTED,
                                                        UmiCmd.UMI_REQ_ATOMIC],
                                            bytes_per_tx=64)

    packets_resp = gen_mergeable_umi_packets(num_bytes=65536,
                                             size=0,
                                             opcode_arr=[UmiCmd.UMI_RESP_READ,
                                                         UmiCmd.UMI_RESP_WRITE],
                                             bytes_per_tx=64)

    # instantiate TX and RX queues.  note that these can be instantiated without
    # specifying a URI, in which case the URI can be specified later via the
    # "init" method

    req = UmiTxRx("umi_req_in.q", "umi_req_out.q", fresh=True)
    resp = UmiTxRx("umi_resp_in.q", "umi_resp_out.q", fresh=True)

    print("### Starting test ###")

    procs = []

    procs.append(multiprocessing.Process(target=umi_loopback, args=(req, packets_req)))

    procs.append(multiprocessing.Process(target=umi_loopback, args=(resp, packets_resp)))

    for proc in procs:
        proc.start()

    for proc in procs:
        proc.join()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (C) 2023 Zero ASIC
# This code is licensed under Apache License 2.0 (see LICENSE for details)

import pytest
import numpy as np
from switchboard import UmiTxRx


def test_regif(sumi_dut, random_seed, sb_umi_valid_mode, sb_umi_ready_mode):

    np.random.seed(random_seed)

    # launch the simulation
    sumi_dut.simulate(plusargs=[('valid_mode', sb_umi_valid_mode),
                                ('ready_mode', sb_umi_ready_mode)])

    # instantiate TX and RX queues.  note that these can be instantiated without
    # specifying a URI, in which case the URI can be specified later via the
    # "init" method

    host = UmiTxRx("host2dut_0.q", "dut2host_0.q", fresh=True)

    print("### Starting test ###")

    # regif accesses are all 32b wide and aligned
    for _ in range(10):
        addr = np.random.randint(0, 16) * 4
        # length should not cross the DW boundary - umi_mem_agent limitation
        data = np.random.randint(2**32, dtype=np.uint32)

        print(f"umi writing 0x{data:08x} to addr 0x{addr:08x}")
        host.write(addr, data)
        print(f"umi read from addr 0x{addr:08x}")
        val = host.read(addr, np.uint32)
        if not (val == data):
            print(f"ERROR umi read from addr 0x{addr:08x} expected {data} actual {val}")
            assert (val == data)


if __name__ == '__main__':
    pytest.main(['-s', '-q', __file__])



#!/usr/bin/env python3

# Copyright (C) 2024 Zero ASIC
# This code is licensed under Apache License 2.0 (see LICENSE for details)

import pytest
import multiprocessing
from switchboard import UmiTxRx, random_umi_packet, delete_queue


def test_mux(sumi_dut, umi_send, sb_umi_valid_mode, sb_umi_ready_mode):
    n = 1000  # Number of transactions to be sent to each mux input port
    in_ports = 4  # Number of input ports. Must match testbench
    out_ports = 1  # Number of output ports. Fixed to 1 for mux

    for x in range(in_ports):
        delete_queue(f'client2rtl_{x}.q')
        delete_queue(f'tee_{x}.q')

    # Instantiate TX and RX queues
    umi = [UmiTxRx('', f'rtl2client_{x}.q', fresh=True) for x in range(out_ports)]
    tee = [UmiTxRx('', f'tee_{x}.q') for x in range(in_ports)]

    # launch the simulation
    sumi_dut.simulate(
            plusargs=[('valid_mode', sb_umi_valid_mode),
                      ('ready_mode', sb_umi_ready_mode)])

    print("### Starting test ###")

    send_procs = []

    for x in range(in_ports):
        send_procs.append(multiprocessing.Process(target=umi_send,
                                                  args=(x, n, out_ports,)))

    for proc in send_procs:
        proc.start()

    recv_queue = [[] for i in range(in_ports)]
    send_queue = [[] for i in range(in_ports)]

    nrecv = 0
    nsend = 0
    while (nrecv < in_ports*n) or (nsend < in_ports*n):
        for i in range(out_ports):
            rxp = umi[i].recv(blocking=False)
            if rxp is not None:
                if nrecv >= in_ports*n:
                    print(f'Unexpected packet received {nrecv}')
                    raise Exception(f'Unexpected packet received {nrecv}')
                else:
                    recv_src = (rxp.srcaddr >> 40)
                    print(f"port {i} receiving srcaddr: 0x{rxp.srcaddr:08x} "
                          f"dstaddr: 0x{rxp.dstaddr:08x} src: {recv_src} #{nrecv}")
                    recv_queue[recv_src].append(rxp)
                    nrecv += 1

        for i in range(in_ports):
            txp = tee[i].recv(blocking=False)
            if txp is not None:
                if nsend >= in_ports*n:
                    raise Exception('Unexpected packet sent')
                else:
                    send_queue[i].append(txp)
                    nsend += 1

    # join Tx senders
    for proc in send_procs:
        proc.join()

    for i in range(in_ports):
        if len(send_queue[i]) != len(recv_queue[i]):
            print(f"packets sent: {len(send_queue[i])} packets received: {len(recv_queue[i])}")
        assert len(send_queue[i]) == len(recv_queue[i])
        for txp, rxp in zip(send_queue[i], recv_queue[i]):
            assert txp == rxp
        print(f"Received {len(recv_queue[i])} packets at port {i}")


@pytest.mark.skip(reason="Must only be run when evaluating performance using waveforms")
def test_round_robin_arb(sumi_dut):
    '''
    This test is used to get an initial idea of the performance impact of
    the arbitration scheme present in the mux. With the thermometer based
    round robin scheme this test shows a performance penalty of up to 4
    cycles for a thermometer masked transaction. This test must be run with
    the waveform enabled.
    '''

    # Instantiate TX and RX queues
    inq = [UmiTxRx(f'client2rtl_{x}.q', '', fresh=True) for x in range(2)]

    # launch the simulation
    sumi_dut.simulate(
            plusargs=[('valid_mode', 1),
                      ('ready_mode', 1)])

    txp = random_umi_packet()
    print(f"Sending cmd: 0x{txp.cmd:08x} "
          f"srcaddr: 0x{txp.srcaddr:08x} dstaddr: 0x{txp.dstaddr:08x}")
    # send the packet to both simulation and local queues
    inq[0].send(txp)
    inq[1].send(txp)
    inq[0].send(txp)


if __name__ == '__main__':
    pytest.main(['-s', '-q', __file__])



from umi.common import UMI
from umi.adapters.common import TileLink
from umi.sumi import FifoFlex
from umi.sumi import Unpack
from umi.sumi import Pack


class UMI2TL(UMI):
    def __init__(self):
        super().__init__('umi2tl',
                         files=['rtl/umi2tl.v'],
                         idirs=['rtl'],
                         deps=[TileLink(),
                               FifoFlex(),
                               Pack(),
                               Unpack()])


if __name__ == "__main__":
    d = UMI2TL()
    d.write_fileset(f"{d.name}.f", fileset="rtl")



from umi.common import UMI
import lambdalib as ll


class Isolate(UMI):
    def __init__(self):
        name = 'umi_isolate'
        super().__init__(name,
                         files=[f'rtl/{name}.v'],
                         deps=[ll.auxlib.Isolo()])


if __name__ == "__main__":
    d = Isolate()
    d.write_fileset(f"{d.name}.f", fileset="rtl")



import pytest
import os


@pytest.fixture(autouse=True)
def test_wrapper(tmp_path):
    '''
    Fixture that automatically runs each test in a test-specific temporary
    directory to avoid clutter.
    '''
    topdir = os.getcwd()
    os.chdir(tmp_path)
    # Run the test.
    yield
    os.chdir(topdir)



from umi.common import UMI
from umi.sumi.umi_endpoint.umi_endpoint import Endpoint
from umi.sumi.umi_fifoflex.umi_fifoflex import FifoFlex
from lambdalib.ramlib import Spram


class MemAgent(UMI):
    def __init__(self):
        super().__init__('umi_memagent',
                         files=['rtl/umi_memagent.v'],
                         deps=[FifoFlex(),
                               Spram(),
                               Endpoint()])


if __name__ == "__main__":
    d = MemAgent()
    d.write_fileset(f"{d.name}.f", fileset="rtl")



#!/usr/bin/env python3

# Copyright (C) 2023 Zero ASIC

import random
from switchboard import SbDut, UmiTxRx, random_umi_packet
from umi import sumi


def main():

    extra_args = {
        '--vldmode': dict(type=int, default=1, help='Valid mode'),
        '--rdymode': dict(type=int, default=1, help='Ready mode'),
        '-n': dict(type=int, default=10, help='Number of transactions'
                   'to send during the test.')
    }

    dut = SbDut('testbench', cmdline=True, extra_args=extra_args,
                trace=False, trace_type='fst', default_main=True)

    # Set up inputs
    dut.input('utils/testbench/testbench_umi_address_remap.v', package='umi')

    dut.use(sumi)

    # Verilator configuration
    dut.set('tool', 'verilator', 'task', 'compile',
            'file', 'config', 'utils/testbench/config.vlt', package='umi')

    # Build simulator
    dut.build()

    # launch the simulation
    dut.simulate(plusargs=[('valid_mode', dut.args.vldmode),
                           ('ready_mode', dut.args.rdymode)])

    # instantiate TX and RX queues.  note that these can be instantiated without
    # specifying a URI, in which case the URI can be specified later via the
    # "init" method

    umi = UmiTxRx("client2rtl_0.q", "rtl2client_0.q", fresh=True)

    print("### Starting random test ###")

    n_sent = 0
    n_recv = 0
    txq = []

    while (n_sent < dut.args.n) or (n_recv < dut.args.n):
        addr = random.randrange(0x0000_0000_0000_0000, 0x0000_07FF_FFFF_FFFF)
        addr = addr & 0xFFFF_FF00_0000_FFF0  # Allow different devices but reduce address space per device

        txp = random_umi_packet(dstaddr=addr, srcaddr=0x0000110000000000)
        if n_sent < dut.args.n:
            if umi.send(txp, blocking=False):
                print(f"Transaction sent: {n_sent}")
                print(str(txp))
                txq.append(txp)
                # Offset
                if ((addr >= 0x0000_0600_0000_0080) and
                        (addr <= 0x0000_06FF_FFFF_FFFF)):
                    addr = addr - 0x0000_0000_0000_0080
                    txq[-1].dstaddr = addr
                # Remap
                elif ((addr & 0xFFFF_FF00_0000_0000) != 0x0000_0400_0000_0000):
                    addr = addr ^ 0x00FF_FF00_0000_0000
                    txq[-1].dstaddr = addr
                n_sent += 1

        if n_recv < dut.args.n:
            rxp = umi.recv(blocking=False)
            if rxp is not None:
                print(f"Transaction received: {n_recv}")
                print(str(rxp))
                if rxp != txq[0]:
                    raise Exception(f'Mismatch! {n_recv}')
                else:
                    txq.pop(0)
                    n_recv += 1


if __name__ == '__main__':
    main()



import inspect
from pathlib import Path
from typing import List, Tuple
from siliconcompiler import Design


class UMI(Design):
    def __init__(self,
                 name: str,
                 files: List[str] = None,
                 idirs: List[str] = None,
                 deps: List[Design] = None,
                 defines: List[str] = None,
                 undefines: List[str] = None,
                 params: List[Tuple] = None):

        super().__init__(name)

        # move this
        cls = self.__class__
        module = inspect.getmodule(cls)
        localpath = Path(inspect.getfile(module)).resolve().parent
        globalpath = Path(__file__).resolve().parent

        # Taking care of Nones
        if idirs is None:
            idirs = []
        if deps is None:
            deps = []
        if defines is None:
            defines = []
        if undefines is None:
            undefines = []
        if params is None:
            params = []

        # dataroot
        self.set_dataroot('localroot', localpath)

        # Setting RTL list, others outside
        with self.active_fileset('rtl'):
            self.set_topmodule(name)
            self.add_idir(globalpath / 'sumi' / 'include')
            for item in files:
                self.add_file(item)
            for item in idirs:
                self.add_idir(item)
            for item in deps:
                self.add_depfileset(item)
            for item in defines:
                self.add_define(item)
            for item in undefines:
                self.add_undefine(item)
            for item in params:
                self.add_param(item[0], item[1])


##################################################
# Standard Definition
##################################################
class Standard(Design):
    def __init__(self):
        super().__init__("umi_standard")
        base_dir = Path(__file__).resolve().parent
        with self.active_fileset("rtl"):
            self.add_idir(base_dir / 'sumi' / 'include')



from umi.common import UMI
from umi.sumi.umi_decode.umi_decode import Decode


class Pack(UMI):
    def __init__(self):
        super().__init__('umi_pack',
                         files=['rtl/umi_pack.v'],
                         deps=[Decode()])


if __name__ == "__main__":
    d = Pack()
    d.write_fileset(f"{d.name}.f", fileset="rtl")



#!/usr/bin/env python3

# Example illustrating how UMI packets handled in the Switchboard Python binding
# Copyright (C) 2023 Zero ASIC
# This code is licensed under Apache License 2.0 (see LICENSE for details)

import time
import numpy as np
from switchboard import UmiTxRx
import pytest


def test_lumi_rnd(lumi_dut, chip_topo, random_seed, sb_umi_valid_mode, sb_umi_ready_mode):

    np.random.seed(random_seed)

    hostdly = np.random.randint(500)
    devdly = np.random.randint(500)
    topo = chip_topo

    # launch the simulation
    lumi_dut.simulate(
        plusargs=[
            ('valid_mode', sb_umi_valid_mode),
            ('ready_mode', sb_umi_ready_mode),
            ('hostdly', hostdly),
            ('devdly', devdly)
        ]
    )

    # instantiate TX and RX queues.  note that these can be instantiated without
    # specifying a URI, in which case the URI can be specified later via the
    # "init" method

    sb = UmiTxRx("sb2dut_0.q", "dut2sb_0.q", fresh=True)
    host = UmiTxRx("host2dut_0.q", "dut2host_0.q", fresh=True)

    print("### Side Band loc reset ###")
    sb.write(0x7000000C, np.uint32(0x00000000), posted=True)

    # Need to add some delay are reassertion before sending things
    # over serial link
    print("### Read local reset ###")
    val32 = sb.read(0x70000000, np.uint32)
    print(f"Read: 0x{val32:08x}")
    assert val32 == 0x00000000

    if topo == '2d':
        width = np.uint32(0x00000000)
        crdt = np.uint32(0x001A001A)
    if topo == '3d':
        width = np.uint32(0x00030000)
        crdt = np.uint32(0x00070007)

        linkactive = 0
        while (linkactive == 0):
            print("### Wait for linkactive ###")
            linkactive = sb.read(0x70000004, np.uint32)
            print(f"Read: 0x{val32:08x}")

        linkactive = 0
        while (linkactive == 0):
            print("### Wait for linkactive ###")
            linkactive = sb.read(0x60000004, np.uint32)
            print(f"Read: 0x{val32:08x}")

        print("### disable Tx ###")
        sb.write(0x60000010, np.uint32(0x0), posted=True)
        sb.write(0x70000010, np.uint32(0x0), posted=True)

        time.sleep(0.1)

        print("### disable Rx ###")
        sb.write(0x70000014, np.uint32(0x0), posted=True)
        sb.write(0x60000014, np.uint32(0x0), posted=True)

        print("### configure loc Rx width ###")
        sb.write(0x70000010, width, posted=True)

        print("### configure rmt Rx width ###")
        sb.write(0x60000010, width, posted=True)

        print("### configure loc Tx width ###")
        sb.write(0x70000014, width, posted=True)

        print("### configure rmt Tx width ###")
        sb.write(0x60000014, width, posted=True)

        print("### Tx init credit ###")
        sb.write(0x60000020, crdt, posted=True)

        print("### Tx init credit ###")
        sb.write(0x70000020, crdt, posted=True)

        print("### Rx enable local ###")
        sb.write(0x70000014, np.uint32(0x1) + width, posted=True)

        print("### Rx enable remote ###")
        sb.write(0x60000014, np.uint32(0x1) + width, posted=True)

        print("### Tx enable remote ###")
        sb.write(0x60000010, np.uint32(0x1) + width, posted=True)

        print("### Tx enable local ###")
        sb.write(0x70000010, np.uint32(0x1) + width, posted=True)

        print("### Tx enable credit ###")
        sb.write(0x60000010, np.uint32(0x11) + width, posted=True)

        print("### Tx enable credit ###")
        sb.write(0x70000010, np.uint32(0x11) + width, posted=True)

    print("### Read loc Rx ctrl ###")
    val32 = sb.read(0x70000014, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### Read loc Tx ctrl ###")
    val32 = sb.read(0x70000010, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### Read rmt Rx ctrl ###")
    val32 = sb.read(0x60000014, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### Read rmt Tx ctrl ###")
    val32 = sb.read(0x60000010, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### UMI WRITE/READ ###")

    for count in range(100):
        # length should not cross the DW boundary - umi_mem_agent limitation
        length = np.random.randint(0, 511)
        dst_addr = 32*np.random.randint(2**(10-5))  # sb limitation - should align to bus width
        src_addr = 32*np.random.randint(2**(10-5))
        data8 = np.random.randint(0, 255, size=length, dtype=np.uint8)
        print(f"umi writing {length+1} bytes to addr 0x{dst_addr:08x}")
        host.write(dst_addr, data8, srcaddr=src_addr)
        print(f"umi read from addr 0x{dst_addr:08x}")
        val8 = host.read(dst_addr, length, np.uint8, srcaddr=src_addr)
        if ~((val8 == data8).all()):
            print(f"ERROR umi read from addr 0x{dst_addr:08x}")
            print(f"Expected: {data8}")
            print(f"Actual: {val8}")
            assert (val8 == data8).all()

    print("### Read loc Tx req credit unavailable ###")
    val32 = sb.read(0x70000030, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### Read loc Tx resp credit unavailable ###")
    val32 = sb.read(0x70000034, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### Read loc Tx req credit available ###")
    val32 = sb.read(0x70000038, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### Read loc Tx resp credit available ###")
    val32 = sb.read(0x7000003C, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### Read rmt Tx req credit unavailable ###")
    val32 = sb.read(0x60000030, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### Read rmt Tx resp credit unavailable ###")
    val32 = sb.read(0x60000034, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### Read rmt Tx req credit available ###")
    val32 = sb.read(0x60000038, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### Read rmt Tx resp credit available ###")
    val32 = sb.read(0x6000003C, np.uint32)
    print(f"Read: 0x{val32:08x}")

    print("### TEST PASS ###")


if __name__ == '__main__':
    pytest.main(['-s', '-q', __file__])



import pytest
import umi


@pytest.mark.parametrize("name", umi.sumi.__all__)
def test_setup_sumi(name):
    assert getattr(umi.sumi, name)().check_filepaths()


@pytest.mark.parametrize("name", umi.lumi.__all__)
def test_setup_lumi(name):
    assert getattr(umi.lumi, name)().check_filepaths()


@pytest.mark.parametrize("name", umi.adapters.__all__)
def test_setup_adapters(name):
    assert getattr(umi.adapters, name)().check_filepaths()



#!/usr/bin/env python3

# Copyright (C) 2023 Zero ASIC
# This code is licensed under Apache License 2.0 (see LICENSE for details)

import random
import numpy as np
from switchboard import SbDut, UmiTxRx
from umi import sumi


def main():

    extra_args = {
        '--vldmode': dict(type=int, default=1, help='Valid mode'),
        '--rdymode': dict(type=int, default=1, help='Ready mode'),
        '-n': dict(type=int, default=10, help='Number of transactions'
                   'to send during the test.')
    }

    dut = SbDut('testbench', cmdline=True, extra_args=extra_args, trace=False, default_main=True)

    # Set up inputs
    dut.input('utils/testbench/testbench_umi2apb.sv', package='umi')

    dut.use(sumi)

    # Verilator configuration
    dut.set('tool', 'verilator', 'task', 'compile', 'file', 'config', 'utils/testbench/config.vlt',
            package='umi')

    dut.build()

    # launch the simulation
    dut.simulate(
        plusargs=[
            ('valid_mode', dut.args.vldmode),
            ('ready_mode', dut.args.rdymode)
        ]
    )

    # instantiate TX and RX queues.  note that these can be instantiated without
    # specifying a URI, in which case the URI can be specified later via the
    # "init" method

    host = UmiTxRx("host2dut_0.q", "dut2host_0.q", fresh=True)

    print("### Starting test ###")

    # regif accesses are all 32b wide and aligned
    for _ in range(dut.args.n):
        addr = np.random.randint(0, 512) * 4
        # length should not cross the DW boundary - umi_mem_agent limitation
        data = np.uint32(random.randrange(2**32-1))

        print(f"umi writing 0x{data:08x} to addr 0x{addr:08x}")
        host.write(addr, data)

        print(f"umi read from addr 0x{addr:08x}")
        val = host.read(addr, np.uint32)
        if not (val == data):
            print(f"ERROR umi read from addr 0x{addr:08x} expected {data} actual {val}")
            assert (val == data)

    print("### TEST PASS ###")


if __name__ == '__main__':
    main()



from umi.common import UMI
from umi.sumi import Unpack
from umi.sumi import Pack


class UMI2APB(UMI):
    def __init__(self):
        super().__init__('umi2apb',
                         files=['rtl/umi2apb.v'],
                         idirs=['rtl'],
                         deps=[Unpack(),
                               Pack()])


if __name__ == "__main__":
    d = UMI2APB()
    d.write_fileset(f"{d.name}.f", fileset="rtl")



#!/usr/bin/env python3

# Copyright (C) 2023 Zero ASIC

import random
import numpy as np
from pathlib import Path
from switchboard import SbDut, UmiTxRx
from siliconcompiler.package import path as sc_path
from umi import sumi


def main():

    extra_args = {
        '--vldmode': dict(type=int, default=1, help='Valid mode'),
        '-n': dict(type=int, default=10, help='Number of transactions'
                   'to send during the test.')
    }

    dut = SbDut('testbench', cmdline=True, extra_args=extra_args,
                trace=False, trace_type='fst', default_main=False)

    # Set up inputs
    dut.input('utils/testbench/testbench_umi2tl_np.v', package='umi')
    dut.input('utils/testbench/testbench_umi2tl_np.cc', package='umi')
    dut.input('utils/testbench/tlmemsim.cpp', package='umi')

    dut.use(sumi)

    # Verilator configuration
    dut.add('tool', 'verilator', 'task', 'compile', 'option', '--coverage')
    header_files_dir = Path(sc_path(dut, 'umi')) / 'utils' / 'testbench'
    dut.set('tool', 'verilator', 'task', 'compile', 'var', 'cflags', f'-I{header_files_dir}')
    dut.set('tool', 'verilator', 'task', 'compile', 'file', 'config', 'utils/testbench/config.vlt', package='umi')

    # Build simulator
    dut.build()

    # launch the simulation
    dut.simulate(plusargs=[('valid_mode', dut.args.vldmode)])

    # instantiate TX and RX queues.  note that these can be instantiated without
    # specifying a URI, in which case the URI can be specified later via the
    # "init" method

    umi = UmiTxRx("client2rtl_0.q", "rtl2client_0.q", fresh=True)

    print("### Starting random test ###")

    n_sent = 0

    while (n_sent < dut.args.n):
        print(f"Transaction {n_sent}:")
        addr = random.randrange(511)
        length = random.choice([1, 2, 4, 8])

        # FIXME: Align address. Limitation of umi2tl converter. Will be fixed in the next release
        addr = addr & (0xFFFFFFF8 | (8-length))

        data8 = np.random.randint(0, 255, size=length, dtype=np.uint8)
        print(f"umi writing {length} bytes:: data: {data8} to addr: 0x{addr:08x}")
        umi.write(addr, data8, srcaddr=0x0000110000000000)
        print(f"umi reading {length} bytes:: from addr 0x{addr:08x}")
        val8 = umi.read(addr, length, np.uint8, srcaddr=0x0000110000000000)
        print(f"umi Read: {val8}")
        if not (val8 == data8).all():
            print(f"ERROR core read from addr 0x{addr:08x} expected {data8} actual {val8}")
        assert (val8 == data8).all()
        n_sent = n_sent + 1


if __name__ == '__main__':
    main()



from umi.common import UMI
from lambdalib.auxlib import Drsync
from umi.sumi import FifoFlex
from umi.sumi import Unpack
from umi.sumi import Pack


class UMI2AXIL(UMI):
    def __init__(self):
        super().__init__('umi2axil',
                         files=['rtl/umi2axil.v'],
                         idirs=['rtl'],
                         deps=[Drsync(),
                               FifoFlex(),
                               Unpack(),
                               Pack()])


if __name__ == "__main__":
    d = UMI2AXIL()
    d.write_fileset(f"{d.name}.f", fileset="rtl")



# Copyright (C) 2023 Zero ASIC
# This code is licensed under Apache License 2.0 (see LICENSE for details)
import os
import platform

import pytest
from siliconcompiler.core import SiliconCompilerError


def setup(chip):
    '''Adds sources/build configuration for the UMI testbench.

    Assumes that user plans to build using sclib verification flow.
    '''
    mydir = os.path.dirname(os.path.abspath(__file__))
    sbdir = os.environ['SBDIR']

    # Add testbench sources
    chip.add('input', 'verilog', f'{mydir}/verilog/umi_testbench.sv')
    chip.add('input', 'verilog', f'{sbdir}/verilog/sim/umi_rx_sim.sv')
    chip.add('input', 'verilog', f'{sbdir}/verilog/sim/umi_tx_sim.sv')

    chip.add('input', 'c', f'{mydir}/cpp/umi_testbench.cc')
    chip.add('input', 'c', f'{sbdir}/dpi/switchboard_dpi.cc')

    # testbench module is top-level
    chip.set('option', 'entrypoint', 'umi_testbench')

    trace = chip.get('option', 'trace')
    if trace:
        chip.add('option', 'define', 'TRACE')

    # Configure tools for verification flow
    cflags = ['-Wno-unknown-warning-option', f'-I{sbdir}/cpp']
    if trace:
        cflags += ['-DTRACE']
    chip.set('tool', 'za_verilator', 'var', 'compile', '0', 'cflags', cflags)
    print('CFLAGS', cflags)

    if platform.system() == 'Darwin':
        cpp_libs = ['-lboost_system', '-lpthread']
    else:
        cpp_libs = ['-lboost_system', '-pthread', '-lrt']
    chip.set('tool', 'za_verilator', 'var', 'compile', '0', 'ldflags', cpp_libs)

    chip.set('tool', 'umidriver', 'path', f'{sbdir}/cpp')

    # Make Verilator allow warnings - missing connections in current testbench
    chip.set('option', 'relax', True)


def compile_tb(chip, module):
    '''Sets up and compiles a testbench from a Chip object.'''
    setup(chip)

    # Set up flow - only run to compile in this method
    chip.set('option', 'flow', 'verification')
    chip.set('option', 'steplist', ['import', 'compile'])
    chip.set('option', 'jobname', 'compile_tb')

    chip.add('option', 'define', f'MOD_UNDER_TEST={module}')

    chip.run()

    # Future runs will reuse the compilation results from this job
    chip.set('option', 'jobinput', 'execute', '0', 'compile_tb')

    return chip


def run_tb(chip, job):
    '''Runs the execute step of the verification flow.

    This function is meant to be run in the context of a pytest test.
    '''
    chip.set('option', 'jobname', job)
    chip.set('option', 'steplist', ['execute'])
    try:
        chip.run()
    except SiliconCompilerError:
        # If step fails, it's probably cause of test failure - no need to print
        # out stacktrace. If in quiet mode, print out contents of umidriver log.
        if chip.get('option', 'quiet'):
            workdir = chip._getworkdir(step='execute', index='0')
            log = f'{workdir}/execute.log'
            with open(log, 'r') as f:
                contents = f.read()
        else:
            contents = ''
        pytest.fail(contents, pytrace=False)



# Design objects
from .axil2umi.axil2umi import AXIL2UMI
from .umi2axil.umi2axil import UMI2AXIL
from .umi2apb.umi2apb import UMI2APB
from .tl2umi.tl2umi import TL2UMI
from .umi2tl.umi2tl import UMI2TL

__all__ = ['AXIL2UMI',
           'TL2UMI',
           'UMI2APB',
           'UMI2AXIL',
           'UMI2TL']



from umi.common import UMI
from umi.sumi.umi_mux.umi_mux import Mux
from umi.sumi.umi_memagent.umi_memagent import MemAgent


class RAM(UMI):
    def __init__(self):
        super().__init__('umi_ram',
                         files=['rtl/umi_ram.v'],
                         deps=[Mux(),
                               MemAgent()])


if __name__ == "__main__":
    d = RAM()
    d.write_fileset(f"{d.name}.f", fileset="rtl")



#!/usr/bin/env python3

# Copyright (C) 2024 Zero ASIC
# This code is licensed under Apache License 2.0 (see LICENSE for details)

import pytest
import random
import numpy as np
from switchboard import UmiTxRx


def test_fifo_flex(sumi_dut, random_seed, sb_umi_valid_mode, sb_umi_ready_mode):

    random.seed(random_seed)
    np.random.seed(random_seed)

    # launch the simulation
    sumi_dut.simulate(plusargs=[('valid_mode', sb_umi_valid_mode), ('ready_mode', sb_umi_ready_mode)])

    # instantiate TX and RX queues.  note that these can be instantiated without
    # specifying a URI, in which case the URI can be specified later via the
    # "init" method

    host = UmiTxRx("host2dut_0.q", "dut2host_0.q", fresh=True)

    print("### Starting test ###")

    for count in range(1000):
        # length should not cross the DW boundary - umi_mem_agent limitation
        length = np.random.randint(0, 255)
        dst_addr = 32*random.randrange(2**(10-5)-1)  # sb limitation - should align to bus width
        src_addr = 32*random.randrange(2**(10-5)-1)
        data8 = np.random.randint(0, 255, size=length, dtype=np.uint8)
        print(f"[{count}] umi writing {length} bytes to addr 0x{dst_addr:08x}")
        host.write(dst_addr, data8, srcaddr=src_addr, max_bytes=16)
        print(f"[{count}] umi read from addr 0x{dst_addr:08x}")
        val8 = host.read(dst_addr, length, np.uint8, srcaddr=src_addr, max_bytes=16)
        if ~((val8 == data8).all()):
            print(f"ERROR umi read from addr 0x{dst_addr:08x}")
            print(f"Expected: {data8}")
            print(f"Actual: {val8}")
            assert (val8 == data8).all()


if __name__ == '__main__':
    pytest.main(['-s', '-q', __file__])



#!/usr/bin/env python3

# Copyright (C) 2024 Zero ASIC
# This code is licensed under Apache License 2.0 (see LICENSE for details)

import pytest
import multiprocessing
from switchboard import UmiTxRx, delete_queue


def test_switch(sumi_dut, umi_send, sb_umi_valid_mode, sb_umi_ready_mode):
    n = 1000  # Number of transactions to be sent to each switch input port
    in_ports = 4  # Number of input ports. Must match testbench
    out_ports = 2  # Number of output ports. Must match testbench

    for x in range(in_ports):
        delete_queue(f'client2rtl_{x}.q')
        delete_queue(f'tee_{x}.q')

    # Instantiate TX and RX queues
    umi = [UmiTxRx('', f'rtl2client_{x}.q', fresh=True) for x in range(out_ports)]
    tee = [UmiTxRx('', f'tee_{x}.q') for x in range(in_ports)]

    # launch the simulation
    sumi_dut.simulate(
            plusargs=[('valid_mode', sb_umi_valid_mode),
                      ('ready_mode', sb_umi_ready_mode)])

    print("### Starting test ###")

    procs = []
    for x in range(in_ports):
        procs.append(multiprocessing.Process(target=umi_send,
                                             args=(x, n, out_ports,)))

    for proc in procs:
        proc.start()

    recv_queue = [[[] for i in range(out_ports)] for j in range(in_ports)]
    send_queue = [[[] for i in range(out_ports)] for j in range(in_ports)]

    nrecv = 0
    nsend = 0
    while (nrecv < in_ports*n) or (nsend < in_ports*n):
        for i in range(out_ports):
            rxp = umi[i].recv(blocking=False)
            if rxp is not None:
                if nrecv >= in_ports*n:
                    print(f'Unexpected packet received {nrecv}')
                    raise Exception(f'Unexpected packet received {nrecv}')
                else:
                    recv_src = (rxp.srcaddr >> 40)
                    print(f"port {i} receiving srcaddr: 0x{rxp.srcaddr:08x} "
                          f"dstaddr: 0x{rxp.dstaddr:08x} src: {recv_src} #{nrecv}")
                    recv_queue[recv_src][i].append(rxp)
                    nrecv += 1

        for i in range(in_ports):
            txp = tee[i].recv(blocking=False)
            if txp is not None:
                if nsend >= in_ports*n:
                    raise Exception('Unexpected packet sent')
                else:
                    send_dst = (txp.dstaddr >> 40)
                    send_queue[i][send_dst].append(txp)
                    nsend += 1

    # join running processes
    for proc in procs:
        proc.join()

    for i in range(in_ports):
        for j in range(out_ports):
            if len(send_queue[i][j]) != len(recv_queue[i][j]):
                print(f"packets sent: {len(send_queue[i][j])} packets received: {len(recv_queue[i][j])}")
            assert len(send_queue[i][j]) == len(recv_queue[i][j])
            for txp, rxp in zip(send_queue[i][j], recv_queue[i][j]):
                assert txp == rxp
            print(f"compared {len(recv_queue[i][j])} packets from port {i} to port {j}")


if __name__ == '__main__':
    pytest.main(['-s', '-q', __file__])



import pytest
from switchboard import SbDut, UmiTxRx, random_umi_packet
from pathlib import Path
from umi import sumi
from fasteners import InterProcessLock
import numpy as np


def pytest_collection_modifyitems(items):
    for item in items:
        if "sumi_dut" in getattr(item, "fixturenames", ()):
            item.add_marker("switchboard")
            pass


@pytest.fixture
def build_dir(pytestconfig):
    return pytestconfig.cache.mkdir('sumi_build')


@pytest.fixture
def sumi_dut(build_dir, request):

    extra_args = {
        '--vldmode': dict(type=int, default=1, help='Valid mode'),
        '--rdymode': dict(type=int, default=1, help='Ready mode'),
    }

    dut = SbDut('testbench', cmdline=True, extra_args=extra_args,
                default_main=True)

    dut.use(sumi)

    # Add testbench
    test_file_name = Path(request.fspath).stem
    assert (test_file_name[:5] == 'test_'), "Test file name must start with test_"
    testbench_name = f'sumi/testbench/testbench_{test_file_name[5:]}.sv'
    dut.input(testbench_name, package='umi')

    # TODO: How to add module/testbench specific parameters
    # dut.add('option', 'define', f'SPLIT={int(split)}')

    # Verilator configuration
    dut.set('tool', 'verilator', 'task', 'compile', 'file', 'config', 'sumi/testbench/config.vlt',
            package='umi')

    # Build simulator
    dut.set('option', 'builddir', build_dir / test_file_name)
    with InterProcessLock(build_dir / f'{test_file_name}.lock'):
        # ensure build only happens once
        # https://github.com/pytest-dev/pytest-xdist/blob/v3.6.1/docs/how-to.rst#making-session-scoped-fixtures-execute-only-once
        dut.build(fast=True)

    yield dut

    dut.terminate()


@pytest.fixture
def umi_send(random_seed):

    def setup(host_num, num_packets_to_send, num_out_ports):
        np.random.seed(random_seed)

        umi = UmiTxRx(f'client2rtl_{host_num}.q', '')
        tee = UmiTxRx(f'tee_{host_num}.q', '')

        for count in range(num_packets_to_send):
            dstport = np.random.randint(num_out_ports)
            dstaddr = (2**8)*np.random.randint(2**32) + dstport*(2**40)
            srcaddr = (2**8)*np.random.randint(2**32) + host_num*(2**40)
            txp = random_umi_packet(dstaddr=dstaddr, srcaddr=srcaddr)
            print(f"port {host_num} sending #{count} cmd: 0x{txp.cmd:08x}"
                  f"srcaddr: 0x{srcaddr:08x} dstaddr: 0x{dstaddr:08x} to port {dstport}")
            # send the packet to both simulation and local queues
            umi.send(txp)
            tee.send(txp)

    return setup


@pytest.fixture
def apply_atomic():

    def setup(origdata, atomicdata, operation, maxrange):
        tempval = origdata
        if (operation == 0):
            tempval = origdata + atomicdata
            if (tempval >= maxrange):
                tempval = tempval - maxrange
        elif (operation == 1):
            tempval = origdata & atomicdata
        elif (operation == 2):
            tempval = origdata | atomicdata
        elif (operation == 3):
            tempval = origdata ^ atomicdata
        elif (operation == 4):
            if (origdata & (maxrange >> 1)):
                origdata = int(origdata) - int(maxrange)
            else:
                origdata = int(origdata)
            if (atomicdata & (maxrange >> 1)):
                atomicdata = int(atomicdata) - int(maxrange)
            else:
                atomicdata = int(atomicdata)
            tempval = origdata if (origdata > atomicdata) else atomicdata
        elif (operation == 5):
            if (origdata & (maxrange >> 1)):
                origdata = int(origdata) - int(maxrange)
            else:
                origdata = int(origdata)
            if (atomicdata & (maxrange >> 1)):
                atomicdata = int(atomicdata) - int(maxrange)
            else:
                atomicdata = int(atomicdata)
            tempval = atomicdata if (origdata > atomicdata) else origdata
        elif (operation == 6):
            tempval = origdata if (origdata > atomicdata) else atomicdata
        elif (operation == 7):
            tempval = atomicdata if (origdata > atomicdata) else origdata
        elif (operation == 8):
            tempval = atomicdata
        else:
            tempval = atomicdata

        return tempval

    return setup



import pytest
from switchboard import SbDut
from umi import lumi
from fasteners import InterProcessLock


def pytest_collection_modifyitems(items):
    for item in items:
        if "lumi_dut" in getattr(item, "fixturenames", ()):
            item.add_marker("switchboard")
            pass


@pytest.fixture
def build_dir(pytestconfig):
    return pytestconfig.cache.mkdir('lumi_build')


@pytest.fixture
def lumi_dut(build_dir, request):

    dut = SbDut('testbench', default_main=True, trace=False)

    # dut = SbDut('testbench', cmdline=True,
    #             default_main=True, trace=True, trace_type='vcd')

    dut.use(lumi)

    # Add testbench
    dut.input('lumi/testbench/testbench_lumi.sv', package='umi')

    # Verilator configuration
    dut.set('tool', 'verilator', 'task', 'compile', 'file', 'config', 'lumi/testbench/config.vlt',
            package='umi')

    # Build simulator
    dut.set('option', 'builddir', build_dir / lumi.__name__)
    with InterProcessLock(build_dir / f'{lumi.__name__}.lock'):
        # ensure build only happens once
        # https://github.com/pytest-dev/pytest-xdist/blob/v3.6.1/docs/how-to.rst#making-session-scoped-fixtures-execute-only-once
        dut.build(fast=True)

    yield dut

    dut.terminate()


@pytest.fixture(params=("2d", "3d"))
def chip_topo(request):
    return request.param
